;; Lego Reader
;;
;; Copyright 2022 - By Michael Kohn
;; https://www.mikekohn.net/
;; mike@mikekohn.net
;;
;; Control a stepper motor with an MSP430G2553 using a SparkFun ProDriver
;; (TC78H670FTG). The stepper motor is a Stepperonline Nema 17 1A 12v motor.
;; A Pololu QTRX-MD-08A 8 channel reflectance sensor is connected
;; to the MSP430G2553 over 8 channels of ADC. The system outputs data over
;; UART to a Western Design Center W65C265SXB board.

.include "msp430x2xx.inc"

; 1 / 9600 = 104.17us
; 1 / 2,000,000 MHz = 0.5us

; interrupt count of 208 = 104usA
;   10 interrupts = 1ms
;  481 interrupts = 50ms
; 1923 interrupts = 200ms
; 9615 interrupts = 1s

RAM equ 0x0200
DATA_BITS equ RAM+0
LAST_BYTE equ RAM+16
DATA_PTR  equ RAM+18
DATA      equ RAM+24

STEPPER_MODE0  equ 0x01
STEPPER_MODE1  equ 0x02
STEPPER_MODE2  equ 0x04
STEPPER_MODE3  equ 0x08
STEPPER_EN     equ 0x10
STEPPER_STBY   equ 0x20
STEPPER_UP_DW  equ STEPPER_MODE0
STEPPER_SET_EN equ STEPPER_MODE1
STEPPER_CLK    equ STEPPER_MODE2
STEPPER_CW_CCW equ STEPPER_MODE3
UART_TX        equ 0x40
SPI_DATA       equ 0x40
SPI_CLK        equ 0x80

;  r4 = temp in main
;  r5 =
;  r6 = counter in read_code
;  r7 = uart_send temp
;  r8 = uart_send counter
;  r9 = interrupt count
; r10 =
; r11 =
; r12 =
; r13 =
; r14 =
; r15 = temp in main

.org 0xc000
start:
  ;; Turn off watchdog
  mov.w #(WDTPW|WDTHOLD), &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Setup stack pointer
  mov.w #0x0400, SP

  ;; Set MCLK to 2 MHz with DCO
  mov.b #DCO_3, &DCOCTL
  mov.b #RSEL_9, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Setup output pins
  ;; P1.0 = Analog Data 0
  ;; P1.1 = Analog Data 1
  ;; P1.2 = Analog Data 2
  ;; P1.3 = Analog Data 3
  ;; P1.4 = Analog Data 4
  ;; P1.5 = Analog Data 5
  ;; P1.6 = Analog Data 6
  ;; P1.7 = Analog Data 7
  ;; P2.0 = MODE3
  ;; P2.1 = MODE2
  ;; P2.2 = MODE1
  ;; P2.3 = MODE0
  ;; P2.4 = EN
  ;; P2.5 = STBY
  ;; P2.6 = TX
  mov.b #0x00, &P1DIR
  mov.b #0x00, &P1OUT
  mov.b #0xff, &P2DIR
  ;mov.b #UART_TX, &P2OUT
  mov.b #0x00, &P2OUT
  mov.b #0x00, &P2SEL

  ;; Setup Timer A
  ;mov.w #216, &TACCR0
  ;mov.w #232, &TACCR0
  mov.w #242, &TACCR0
  mov.w #TASSEL_2|MC_1, &TACTL ; SMCLK, DIV1, COUNT to TACCR0
  mov.w #CCIE, &TACCTL0
  mov.w #0, &TACCTL1

  ;; Setup UART (9600 @ 2 MHz)
  ;mov.b #UCSSEL_2|UCSWRST, &UCA0CTL1
  ;mov.b #0, &UCA0CTL0
  ;mov.b #UCBRS_2, &UCA0MCTL
  ;mov.b #216, &UCA0BR0
  ;mov.b #0, &UCA0BR1
  ;bic.b #UCSWRST, &UCA0CTL1

  ;; Set up A/D. Enable all 8 channels.
  mov.w #ADC10SHT_3|ADC10ON, &ADC10CTL0
  mov.w #INCH_0|ADC10SSEL_2, &ADC10CTL1
  mov.b #0xff, &ADC10AE0

  ;; memset(RAM, 0, 0x40)
  mov.w #RAM, r15
memset:
  mov.w #0, @r15
  add.w #2, r15
  cmp.w #RAM + 0x40, r15
  jnz memset

  mov.w #0, r9
  mov.w #DATA, &DATA_PTR
  mov.w #DATA, &DATA_PTR+2

  ;; Turn on interrupts
  eint

  call #step_init
  call #delay_5s
  call #read_code

  ;mov.w #100, r15
  ;call #step_forward
  ;bic.b #STEPPER_EN, &P2OUT

main:
  jmp main

;; read_code()
read_code:
  mov.w #19, r6
read_code_loop:
  call #read_sensor_byte
  call #spi_send
  mov.w #100, r15
  call #step_forward
  call #delay_200ms
  dec.w r6
  jnz read_code_loop
  bic.b #STEPPER_EN, &P2OUT
  mov.b #0, r15
  call #spi_send
  ret

;; spi_send(r15)
spi_send:
  mov.b #8, r8
spi_send_next:
  bit.b #0x80, r15
  jnz spi_send_data_not_0
  bic.b #SPI_DATA, &P2OUT
spi_send_data_not_0:
  bit.b #0x80, r15
  jz spi_send_data_not_1
  bis.b #SPI_DATA, &P2OUT
spi_send_data_not_1:
  bis.b #SPI_CLK, &P2OUT
  call #spi_clock_delay
  bic.b #SPI_CLK|SPI_DATA, &P2OUT
  call #spi_clock_delay
  rla.w r15
  dec.b r8
  jnz spi_send_next
  ret

spi_clock_delay:
  mov.b #30, r14
spi_clock_delay_loop:
  dec.b r14
  jnz spi_clock_delay_loop
  ret

;; uart_send(r15)
uart_send:
  ;; Add start and stop bit.
  bis.w #0x0100, r15
  rla.w r15
  mov.w #10, r8
  ;; Sync the timer interrupt.
  mov.w r9, r7
uart_send_sync:
  cmp.w r9, r7
  jz uart_send_sync
  ;; Start bit, 8 bits of data, stop bit.
uart_send_data:
  mov.w r9, r7
  bit.b #0x01, r15
  jnz uart_send_data_not_0
  bic.b #UART_TX, &P2OUT
uart_send_data_not_0:
  bit.b #0x01, r15
  jz uart_send_data_not_1
  bis.b #UART_TX, &P2OUT
uart_send_data_not_1:
  rra.w r15
uart_send_wait_data_bit:
  cmp.w r9, r7
  jz uart_send_wait_data_bit
  dec.b r8
  jnz uart_send_data
  ret

;; step_init()
step_init:
  bic.b #0x3f, &P2OUT
  bis.b #STEPPER_MODE3 | STEPPER_MODE2, &P2OUT
  call #delay_1ms
  bis.b #STEPPER_STBY, &P2OUT
  call #delay_200us
  bic.b #0x0f, &P2OUT
  bis.b #STEPPER_EN, &P2OUT
  ret

;; step_forward(r15)
step_forward:
  bis.b #STEPPER_CLK, &P2OUT
  call #delay_1ms
  bic.b #STEPPER_CLK, &P2OUT
  call #delay_1ms
  dec.w r15
  jnz step_forward
  ret

;; read_sensor_byte() : r15
read_sensor_byte:
  mov.b #0x00, r15
  mov.w #0x0000, r4
  mov.w #DATA_BITS, r5
read_sensor_byte_next_bit:
  ;; Start new ADC encoding.
  mov.w #ADC10SSEL_2, r14
  bis.w r4, r14
  mov.w #0, &ADC10CTL0
  mov.w r14, &ADC10CTL1
  mov.w #ENC|ADC10SHT_3|ADC10ON|ADC10SC, &ADC10CTL0
read_sensor_byte_wait_for_data:
  bit.w #ADC10BUSY, &ADC10CTL1
  jnz read_sensor_byte_wait_for_data
  ;; Read 10 bit DAC and store in memory (more for debugging).
  mov.w &ADC10MEM, r14
  mov.w r14, @r5
  add.w #2, r5
  ;; Shift in bit.
  rla.w r15
  cmp.w #0x2ff, r14
  jlo read_sensor_byte_not_1
  bis.w #1, r15
read_sensor_byte_not_1:
  ;; Next input..
  add.w #0x1000, r4
  cmp.w #0x8000, r4
  jnz read_sensor_byte_next_bit
  mov.w r15, &LAST_BYTE
  mov.w &DATA_PTR, r14
  mov.b r15, @r14
  add.w #1, &DATA_PTR
  ret

test_uart_time:
  mov.w r9, r7
test_uart_time_sync:
  cmp.w r9, r7
  jz test_uart_time_sync
  mov.w #0, r9
  bic.b #UART_TX, &P2OUT
test_uart_time_wait:
  cmp.w #10, r9
  jnz test_uart_time_wait
  bis.b #UART_TX, &P2OUT
  ret

delay_1ms:
  mov.w #0, r9
delay_1ms_loop:
  cmp.w #10, r9
  jnz delay_1ms_loop
  ret

delay_200us:
  mov.w #0, r9
delay_200us_loop:
  cmp.w #3, r9
  jnz delay_200us_loop
  ret

delay_200ms:
  mov.w #0, r9
delay_200ms_loop:
  cmp.w #1923, r9
  jnz delay_200ms_loop
  ret

delay_5s:
  mov.w #0, r9
delay_5s_loop:
  cmp.w #48075, r9
  jnz delay_5s_loop
  ret

timer_interrupt:
  inc.w r9
  reti

.org 0xfff2
  dw timer_interrupt       ; Timer0_A3 TACCR0, CCIFG
.org 0xfffe
  dw start                 ; Reset

